/* D1M-WX1_IoT_Delta.ino
   D1 Mini Weather Station (Solar)
   Posts to ThingSpeak only

   05/21/2018 - New release. Updated BME library,
                added Wire.begin() as needed by BME2280
   07/11/2017 - Bethany release

   Set serial monitor to 115,200 baud

   TO DO:
      Add test for BH1750 failure/not found
      Add wifiManager

   Copyright(c) 2018 Berger Engineering dba IoT Kits

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in all
   copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
*/

#include <Wire.h>               // [builtin] I2C bus
#include <BME280I2C.h>          // [manager] Tyler Glenn https://github.com/finitespace/BME280
#include <BH1750.h>             // [add zip] https://github.com/claws/BH1750
#include <ESP8266WiFi.h>        // [builtin] ESP8266 WiFi

#include "config.h"             // Locate your configuration file in same folder as this sketch

// *******************************************************
// ******************* DEFAULTS **************************
// *******************************************************
// DO NOT CHANGE THESE DEFAULTS
const char IOT_SERVER[] = "api.thingspeak.com"; // ThingSpeak Server

// *******************************************************
// ******************* GLOBALS ***************************
// *******************************************************
const float HPA_TO_INHG = 0.0295299830714;      // hPa (mb) to inHg pressure
String unitStatus = "";                         // for weather station status

// structure to hold sensor measurements in metric units
struct sensorData {
  float stationPressure;          // measured station pressure in hPa (mb)
  float seaLevelPressure;         // calculated SLP
  float temperature;              // degrees Celsius
  float humidity;                 // relative humidity %
  unsigned int lightIntensity;    // lux| Sketchboo
  float cellVoltage;              // volts
  long wifiRSSI;                  // dBm
} stationData;                    // declare struct variable

// *******************************************************
// ********* INSTANTIATE OBJECTS *************************
// *******************************************************
BME280I2C myBME280;     // barometric pressure / temperature / humidity sensor
BH1750 myBH1750;        // light intensity sensor
WiFiClient client;      // WiFi connection

// *******************************************************
// ********************* SETUP ***************************
// *******************************************************
void setup() {
  Serial.begin(115200);             // initialize the serial port
  pinMode(LED_BUILTIN, OUTPUT);     // set builtin LED for output
  Wire.begin();                     // wake up the I2C bus - BME280 seems to need this
  // initialize BME280 pressure/temperature/humidity sensor
  if ( myBME280.begin() == false) { // initialize sensor & return status
    Serial.println("BME280 sensor not found.");
    unitStatus += "BME280 sensor not found.\n";
    blinkLED(4);                    // notify the user
  } // if BME280

  // initialize BH1750 light sensor
  // the BH1750 library does not have a test for presence
  myBH1750.begin();

  logonToRouter();           // logon to local Wi-Fi

  readSensors();             // read all sensors stationData struct
  printToSerialPort();       // display data to local serial monitor
  postToThingSpeak();        // send data to ThingSpeak

  // all done, now go to sleep
  blinkLED(3);               // notify the user
  enterSleep(SLEEP_INTERVAL);
} //setup()

// *******************************************************
// ******************** LOOP *****************************
// *******************************************************
void loop() {
  // there is nothing to do here
  // everything is done in setup()
} // loop()

// *******************************************************
// ****************** SUPPORTING FUNCTIONS ***************
// *******************************************************

// *******************************************************
// *********** logon to your Wi-Fi router ****************
// *******************************************************
void logonToRouter() {
  String exitMessage = "";
  int count = 0;
  WiFi.begin(WIFI_SSID, WIFI_PASSWORD);
  while ( WiFi.status() != WL_CONNECTED ) {
    count++;
    // give up if more than 15 tries
    if ( count > 15 ) {
      Serial.print("WiFi failed. Restart in 5 minutes. ");
      blinkLED(5);                    // notify the user
      enterSleep(300);                 // retry after 5 minutes
    } // if > 15
    // otherwise if < 15 blink LED and wait 500ms before checking connection
    blinkLED(1); // blink LED on each attempt to connect
    delay(500);  // one-half second delay between checks
    Serial.print(".");
  } // while not connected
  // WiFi is sucesfully connected
  Serial.println("");                 // new line to show IP address
  Serial.print("WiFi connected to IP: ");
  Serial.println(WiFi.localIP().toString());
} // logonToRouter()

// *******************************************************
// ******* read sensors into a sensorData struct *********
// *******************************************************
void readSensors() {
  stationData = {0};    // initialize elements to zero

  // a fudgeFactor corrects for voltage divider component variation
  // as measured by user in the calbration step
  float fudgeFactor = dmmVoltage / adcVoltage;

  int samples = 3;      // number of samples to take for average

  // declare and initialize temporary variables
  float t = 0;          // temperature C
  float h = 0;          // humidity %
  float sp = 0;         // station pressure mb or hPa
  float slp = 0;        // sea level pressure
  unsigned int li = 0;  // light intensity lux
  float cv = 0;         // cell voltage
  long ssi = 0;         // signal strength

  // read sensors multiple times and take average
  for (int i = 0; i < samples; i++) {
    myBME280.read(sp, t, h, BME280::TempUnit_Celsius, BME280::PresUnit_hPa);

    // calculate the Sea Level Pressure from the station pressure
    slp = calculateSeaLevelPressure(t, sp, STATION_ELEV);

    // read light level in lux
    li = myBH1750.readLightLevel();

    // read analog voltage from the Analog to Digital Converter
    // on D1 Mini this is 0 - 1023 for voltages 0 to 3.2V
    // the D1M-WX1 has an external resistor to extend the range to 5.0 Volts
    cv = 5.0 * analogRead(A0) * fudgeFactor / 1023.0;

    // read the Wi-Fi signal strength
    ssi = WiFi.RSSI();

    // accumulate the values of each sensor
    stationData.temperature += t;
    stationData.humidity += h;
    stationData.stationPressure += sp;
    stationData.seaLevelPressure += slp;
    stationData.lightIntensity += li;
    stationData.cellVoltage += cv;
    stationData.wifiRSSI += ssi;

    delay(50);   // provide some delay to let sensors settle
  } // for()

  // divide the accumulated values by the number of samples
  // to get an average
  stationData.temperature /= (float)samples;
  stationData.humidity /= (float)samples;
  stationData.stationPressure /= (float)samples;
  stationData.seaLevelPressure /= (float)samples;
  stationData.lightIntensity /= (long)samples;
  stationData.cellVoltage /= (float)samples;
  stationData.wifiRSSI /= (int)samples;
} // readSensors()

// *******************************************************
// ********* print raw data to the serial port ***********
// *******************************************************
void printToSerialPort() {
  // '\t' is the C++ escape sequence for tab
  // header line
  Serial.println("\t°C(°F)\t\tRH%\tSP mb\tSLP(in)\t\tLux\tVolt");
  // data line
  Serial.print("Data\t");
  Serial.print(stationData.temperature, 1);
  Serial.print("(");
  Serial.print(stationData.temperature * 1.8 + 32, 1);
  Serial.print(")\t");
  Serial.print(stationData.humidity, 1);
  Serial.print("\t");
  Serial.print(stationData.stationPressure, 1);
  Serial.print("\t");
  Serial.print(stationData.seaLevelPressure, 1);
  Serial.print("(");
  Serial.print(stationData.seaLevelPressure * HPA_TO_INHG, 2);
  Serial.print(")\t");
  Serial.print(stationData.lightIntensity);
  Serial.print("\t");
  Serial.println(stationData.cellVoltage, 2);
  Serial.println("----------------------------------------------------------------------------");
} // printToSerialPort()

// *******************************************************
// ************* post to ThingSpeak **********************
// *******************************************************
void postToThingSpeak() {
  // assemble and post the data
  if ( client.connect(IOT_SERVER, 80) == true ) {
    Serial.println("ThingSpeak Server connected.");
    // declare dataString as a String and initialize with the apiWriteKey
    String dataString = apiWriteKey;

    // cocatenate each field onto the end of dataString
    dataString += "&field1=";
    dataString += String(stationData.temperature);  // temp in C
    dataString += "&field2=";
    dataString += String(stationData.humidity);
    dataString += "&field3=";
    dataString += String(stationData.stationPressure);
    dataString += "&field4=";
    dataString += String(stationData.seaLevelPressure);
    dataString += "&field5=";
    dataString += String(stationData.lightIntensity);
    dataString += "&field6=";
    dataString += String(stationData.cellVoltage);
    dataString += "&field7=";
    dataString += String(stationData.wifiRSSI);
    dataString += "&field8=";
    dataString += String( 1.8 * stationData.temperature + 32 ); // temp in F
    dataString += "&status=";
    dataString += unitStatus;
    Serial.print(unitStatus);   // show status on local serial monitor

    // find the number of characters in dataString
    String dataStringLength = String(dataString.length());

    // post the data to ThingSpeak
    client.println("POST /update HTTP/1.1");
    client.println("Host: api.thingspeak.com");
    client.println("Connection: close");
    client.println("X-THINGSPEAKAPIKEY: " + apiWriteKey);
    client.println("Content-Type: application/x-www-form-urlencoded");
    client.println("Content-Length: " + dataStringLength);
    client.println("");
    client.print(dataString);
    Serial.println("ThingSpeak data sent.");
  }
  client.stop();
  Serial.println("ThingSpeak disconnected.");
} // postToThingSpeak()

// *******************************************************
// calculate relative sea level pressure from absolute station pressure in hPa
// temperature in °C, elevation in m
// http://www.sandhurstweather.org.uk/barometric.pdf
// http://keisan.casio.com/exec/system/1224575267
// *******************************************************
float calculateSeaLevelPressure(float celsius, float stationPressure, float elevation) {
  float slP = stationPressure / pow(2.718281828, -(elevation / ((273.15 + celsius) * 29.263)));
  return slP;
} // calculateSeaLevelPressure()

// *******************************************************
// ***************** Net Time Protocol *******************
// *******************************************************
// cribbed from Tom Igoe and others
// added retries after time out to improve reliability KWB 7/8/17

// globals for NTP routines
const int NTP_PACKET_SIZE = 48;         // NTP time is in the first 48 bytes of message
byte packetBuffer[NTP_PACKET_SIZE];     // buffer to hold incoming & outgoing packets

// *******************************************************
// ******************** enterSleep ***********************
// *******************************************************
void enterSleep(long sleep) {
  // sleep is in seconds
  Serial.print("Entering deep sleep for ");
  Serial.print(sleep);
  Serial.println(" seconds.");
  delay(200);                       // delay to let things settle
  // WAKE_RF_DEFAULT wakes the ESP8266 with WiFi enabled
  ESP.deepSleep(sleep * 1000000L, WAKE_RF_DEFAULT);
} // enterSleep()

// *******************************************************
// ******************** Blink LED ************************
// *******************************************************
// this flashes the onboard LED to indicate various internal messages
void blinkLED(int flashes) {
  // set LED_BUILTIN pin to Output mode in setup()
  for (int i = 0; i < flashes; i++) {
    digitalWrite(LED_BUILTIN, LOW);  // Turn the LED *ON*
    delay(20);                       // short flash for low energy consumption
    digitalWrite(LED_BUILTIN, HIGH); // Turn the LED *OFF*
    delay(250);                      // time between flashes
  }
} // blinkLED()

// *******************************************************
// *********************** END ***************************
// *******************************************************
